"""HDF5ExtensionArray module."""
import numpy as np
import h5py
import uuid
from h5pandas.h5datatype import HDF5Dtype
import numbers
from functools import cached_property
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Literal,
    Sequence,
    cast,
)
from pandas._typing import (
    AxisInt,
    Dtype,
    FillnaOptions,
    InterpolateOptions,
    NpDtype,
    Scalar,
    Self,
    npt,
)

import pandas

from pandas import (
    Index,
    Series,
)

from pandas.compat.numpy import function as nv

from pandas._libs import lib
from pandas.core.dtypes.common import (
    is_array_like,
    is_bool_dtype,
    is_integer,
    is_list_like,
    is_object_dtype,
    is_scalar,
)

from pandas.core import (
    arraylike,
    nanops,
    missing,
    roperator,
)
from pandas.core.indexers import (
    check_array_indexer,
    unpack_tuple_and_ellipses,
    validate_indices,
)

from pandas.core.dtypes.generic import (
    ABCDataFrame,
    ABCIndex,
    ABCSeries,
)

from pandas.core.dtypes.missing import isna

from pandas._typing import (
    ArrayLike,
    AxisInt,
    Dtype,
    FillnaOptions,
    Iterator,
    NpDtype,
    PositionalIndexer,
    Scalar,
    SortKind,
    TakeIndexer,
    TimeAmbiguous,
    TimeNonexistent,
    npt,
)

from pandas.core.algorithms import _ensure_arraylike, value_counts_arraylike


class HDF5ExtensionArray( np.lib.mixins.NDArrayOperatorsMixin, pandas.api.extensions.ExtensionArray):

    _HANDLED_TYPES = (np.ndarray, numbers.Number)
    # __pandas_priority__ = 5000
    # __array_priority__ = 1000

    def __init__(self, dataset: h5py.Dataset, column_index: int = 0, dtype=None) -> None:
        """
        Create a HDF5ExtensionArray from a dataset and a column number.

        Parameters
        ----------
        dataset : h5py.Dataset
        column_index : int, optional
            index of the column inside the dataset that will become an HDF5ExtensionArray. The default is 0.
        dtype : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None
            DESCRIPTION.

        """
        if isinstance(dataset, HDF5ExtensionArray):
            dataset = dataset._dataset
        if not isinstance(dataset, (h5py.Dataset)):
            # if dtype is None:
            #     if isinstance(dataset, (list, tuple)):
            #         dataset = np.array()
            #     dtype = dataset.dtype
            if not isinstance(dataset, np.ndarray):
                dataset = np.array(dataset, dtype=dtype)

            if len(dataset):
                # dataset = np.array(dataset, dtype=dtype)
                # dataset = dataset.T
                column_index = 0

                # f = h5py.File("h5pyArray_{}".format(uuid.uuid4()), 'w', libver='latest', driver="core", backing_store=False)
                # try:
                #     dataset = f.create_dataset("column", data=dataset, shape=(len(dataset), 1), maxshape=(None, None), chunks=(len(dataset), 1), dtype=dtype)
                # except OSError:
                #     try:
                #         dataset = f.create_dataset("column", data=dataset, shape=(len(dataset), 1), maxshape=(None, None), chunks=(len(dataset), 1), dtype=h5py.opaque_dtype(dtype))
                #     except TypeError:
                #         dataset = np.array(dataset, dtype=dtype)
                # except TypeError:
                #     dataset = np.array([dataset], dtype=dtype)
                #     dataset = dataset.T

            else:
                dataset = np.empty((0,))
                column_index = 0
        self._dataset = dataset
        self._column_index = column_index
        self._dtype = HDF5Dtype(self._dataset.dtype)

    @classmethod
    def _from_sequence(cls, scalars, *, dtype: np.dtype | None = None, copy: bool = False):
        """
        Construct a new ExtensionArray from a sequence of scalars.

        Parameters
        ----------
        scalars: Sequence
            Each element will be an instance of the scalar type for this
            array, ``cls.dtype.type`` or be converted into this type in this method.
        dtype: dtype, optional
            Construct for this particular dtype. This should be a np.dtype
            compatible with the ExtensionArray.
        copy: bool, default False
            If True, copy the underlying data.

        Returns
        -------
        ExtensionArray
        """
        if dtype is None:
            return HDF5ExtensionArray(np.array(scalars, dtype=dtype))

        if not isinstance(dtype, np.dtype):
            dtype = np.dtype(dtype.type)
        return HDF5ExtensionArray(np.array(scalars, dtype=dtype))

    @classmethod
    def _from_sequence_of_strings(
        cls, strings, *, dtype: np.dtype | None = None, copy: bool = False
    ):
        """
        Construct a new ExtensionArray from a sequence of strings.

        Parameters
        ----------
        strings: Sequence
            Each element will be an instance of the scalar type for this
            array, ``cls.dtype.type``.
        dtype: dtype, optional
            Construct for this particular dtype. This should be a np.dtype
            compatible with the ExtensionArray.
        copy: bool, default False
            If True, copy the underlying data.

        Returns
        -------
        ExtensionArray
        """
        return HDF5ExtensionArray(np.array(strings, dtype=dtype))

    @classmethod
    def _from_factorized(cls, values, original):
        """
        Reconstruct an ExtensionArray after factorization.

        Parameters
        ----------
        values: ndarray
            An integer ndarray with the factorized values.
        original: ExtensionArray
            The original ExtensionArray that factorize was called on.

        See Also
        --------
        factorize: Top-level factorize method that dispatches here.
        ExtensionArray.factorize: Encode the extension array as an enumerated type.
        """

        return HDF5ExtensionArray(values)

    def __getitem__(self, item: PositionalIndexer):
        """
        Select a subset of self.

        Parameters
        ----------
        item: int, slice, or ndarray
            * int: The position in 'self' to get.

            * slice: A slice object, where 'start', 'stop', and 'step' are
              integers or None

            * ndarray: A 1-d boolean NumPy ndarray the same length as 'self'

            * list[int]:  A list of int

        Returns
        -------
        item: scalar or ExtensionArray

        Notes
        -----
        For scalar ``item``, return a scalar value suitable for the array's
        type. This should be an instance of ``self.dtype.type``.

        For slice ``key``, return an instance of ``ExtensionArray``, even
        if the slice is length 0 or 1.

        For a boolean mask, return an instance of ``ExtensionArray``, filtered
        to the values where ``item`` is True.
        """
        item = check_array_indexer(self, item)
        if is_scalar(item):
            try:
                if self._dataset.ndim > 1:
                    return self._dataset[item, self._column_index, ...]
                else:
                    return self._dataset[item]
            except (ValueError, TypeError):
                raise IndexError("only integers, slices (`:`), ellipsis (`...`), numpy.newaxis "
                                 "(`None`) and integer or boolean arrays are valid indices")

        elif isinstance(item, slice) and item == slice(None):
            return HDF5ExtensionArray(self._dataset, self._column_index)
        elif isinstance(item, tuple) and (slice(None) in item or Ellipsis in item):
            if self._dataset.ndim > 1:
                dataset = self._dataset[*item, self._column_index, ...]
            else:
                dataset = self._dataset[*item]
            return HDF5ExtensionArray(dataset)
        else:
            # FIXME : AVOID COPY HERE
            if self._dataset.ndim > 1:
                dataset = self._dataset[item, self._column_index, ...]
            else:
                dataset = self._dataset[item]
            return HDF5ExtensionArray(dataset)

    def __setitem__(self, key, value) -> None:
        """
        Set one or more values inplace.

        This method is not required to satisfy the pandas extension array
        interface.

        Parameters
        ----------
        key: int, ndarray, or slice
            When called from , e.g. ``Series.__setitem__``, ``key`` will be
            one of

            * scalar int
            * ndarray of integers.
            * boolean ndarray
            * slice object

        value: Extensionnp.dtype.type, Sequence[Extensionnp.dtype.type], or object
            value or values to be set of ``key``.

        Returns
        -------
        None
        """
        # Some notes to the ExtensionArray implementor who may have ended up
        # here. While this method is not required for the interface, if you
        # *do* choose to implement __setitem__, then some semantics should be
        # observed:
        #
        # * Setting multiple values : ExtensionArrays should support setting
        #   multiple values at once, 'key' will be a sequence of integers and
        #  'value' will be a same-length sequence.
        #
        # * Broadcasting : For a sequence 'key' and a scalar 'value',
        #   each position in 'key' should be set to 'value'.
        #
        # * Coercion : Most users will expect basic coercion to work. For
        #   example, a string like '2018-01-01' is coerced to a datetime
        #   when setting on a datetime64ns array. In general, if the
        #   __init__ method coerces that value, then so should __setitem__
        # Note, also, that Series/DataFrame.where internally use __setitem__
        # on a copy of the data.
        # FIXME

        if self._dataset.ndim > 1:
            return self._dataset.__setitem__((key, self._column_index, Ellipsis), value)
        else:
            return self._dataset.__setitem__(key, value)

    def __len__(self) -> int:
        """
        Length of this array.

        Returns
        -------
        length: int
        """
        return len(self._dataset)

    def __iter__(self) -> Iterator[Any]:
        """Iterate over elements of the array."""
        # This needs to be implemented so that pandas recognizes extension
        # arrays as list-like. The default implementation makes successive
        # calls to ``__getitem__``, which may be slower than necessary.
        for i in range(len(self)):
            yield self[i]

    def isnan(self) -> np.ndarray:
        return self.isna()

    def isna(self) -> np.ndarray:
        """
        A 1-D array indicating if each value is missing.

        Returns
        -------
        numpy.ndarray or pandas.api.extensions.ExtensionArray
            In most cases, this should return a NumPy ndarray. For
            exceptional cases like ``SparseArray``, where returning
            an ndarray would be expensive, an ExtensionArray may be
            returned.

        Notes
        -----
        If returning an ExtensionArray, then

        * ``na_values._is_boolean`` should be True
        * `na_values` should implement: func: `ExtensionArray._reduce`
        * ``na_values.any`` and ``na_values.all`` should be implemented
        """
        return np.isnan(self._ndarray)

    def __contains__(self, item: object) -> bool | np.bool_:
        """Return for `item in self`."""
        # GH37867
        # comparisons of any item to pd.NA always return pd.NA, so e.g. "a" in [pd.NA]
        # would raise a TypeError. The implementation below works around that.
        if is_scalar(item) and isna(item):
            if not self._can_hold_na:
                return False
            elif item is self.dtype.na_value or isinstance(item, self.dtype.type):
                return self._hasna
            else:
                return False
        else:
            # error: Item "ExtensionArray" of "Union[ExtensionArray, ndarray]" has no
            # attribute "any"
            return (item == self._ndarray).any()  # type: ignore[union-attr]

    @classmethod
    def _concat_same_type(cls, to_concat):
        """
        Concatenate multiple instances of H5pyArray.

        Args:
            to_concat(list): List of H5pyArray instances to concatenate.

        Returns:
            H5pyArray: The concatenated H5pyArray.
        """
        f = h5py.File("h5pyArray_concat{}".format(uuid.uuid4()), 'w', libver='latest', driver="core", backing_store=False, locking=False)
        # TODO : utiliser l'argument out ?

        array = np.concatenate([subdataset._dataset[()] for subdataset in to_concat], axis=0)
        dataset = f.create_dataset("extension", data=array, shape=(len(array), 1), maxshape=(None, None), chunks=(len(array), 1))

        return cls(dataset)

    def view(self, dtype: Dtype | None = None) -> ArrayLike:
        """
        Return a view on the array.

        Parameters
        ----------
        dtype : str, np.dtype, or ExtensionDtype, optional
            Default None.

        Returns
        -------
        ExtensionArray or np.ndarray
            A view on the :class:`ExtensionArray`'s data.

        Examples
        --------
        This gives view on the underlying data of an ``ExtensionArray`` and is not a
        copy. Modifications on either the view or the original ``ExtensionArray``
        will be reflectd on the underlying data:

        >>> arr = pd.array([1, 2, 3])
        >>> arr2 = arr.view()
        >>> arr[0] = 2
        >>> arr2
        <IntegerArray>
        [2, 2, 3]
        Length: 3, dtype: Int64
        """
        # NB:
        # - This must return a *new* object referencing the same data, not self.
        # - The only case that *must* be implemented is with dtype=None,
        #   giving a view with the same dtype as self.
        if dtype is not None:
            raise NotImplementedError(dtype)
        return self[:]

    def to_numpy(
        self,
        dtype: np.dtype | None = None,
        copy: bool = False,
        na_value: object = lib.no_default,
    ) -> np.ndarray:
        """
        Convert to a NumPy ndarray.

        This is similar to: meth: `numpy.asarray`, but may provide additional control
        over how the conversion is done.

        Parameters
        ----------
        dtype: str or numpy.dtype, optional
            The dtype to pass to: meth: `numpy.asarray`.
        copy: bool, default False
            Whether to ensure that the returned value is a not a view on
            another array. Note that ``copy = False`` does not *ensure * that
            ``to_numpy()`` is no-copy. Rather, ``copy = True`` ensure that
            a copy is made, even if not strictly necessary.
        na_value: Any, optional
            The value to use for missing values. The default value depends
            on `dtype` and the type of the array.

        Returns
        -------
        numpy.ndarray
        """
        # We do the copy anyway
        num_array = self._ndarray
        try:
            num_array = num_array.view(dtype)
        except TypeError:
            num_array = num_array.astype(dtype)
        return num_array

    # def __array__(self, *args, **kwargs):
    #     return self._dataset[:, self._column_index, ...]

    # ------------------------------------------------------------------------
    # Required attributes
    # ------------------------------------------------------------------------

    @property
    def dtype(self) -> HDF5Dtype:
        """An instance of 'np.dtype'."""
        return self._dtype

    @property
    def shape(self):
        """Return a tuple of the array dimensions."""
        return (len(self),)

    @property
    def size(self) -> int:
        """The number of elements in the array."""
        # error: Incompatible return value type (got "signedinteger[_64Bit]",
        # expected "int")  [return-value]
        return np.prod(self.shape)  # type: ignore[return-value]

    @property
    def ndim(self) -> int:
        """Extension Arrays are only allowed to be 1-dimensional."""
        return 1

    @property
    def nbytes(self) -> int:
        """The number of bytes needed to store this object in memory."""
        # If this is expensive to compute, return an approximate lower bound
        # on the number of bytes needed.
        return self.__len__()*self.dtype._numpy_dtype.itemsize

    @property
    def _ndarray(self):
        if self._dataset.ndim > 1:
            return self._dataset[:, self._column_index, ...]
        else:
            return self._dataset

    def astype(self, dtype, copy: bool = True) -> ArrayLike:
        """
        Cast to a NumPy array or ExtensionArray with 'dtype'.

        Parameters
        ----------
        dtype: str or dtype
            Typecode or data-type to which the array is cast.
        copy: bool, default True
            Whether to copy the data, even if not necessary. If False,
            a copy is made only if the old dtype does not match the
            new dtype.

        Returns
        -------
        np.ndarray or pandas.api.extensions.ExtensionArray
            An ExtensionArray if dtype is Extensionnp.dtype,
            Otherwise a NumPy ndarray with 'dtype' for its dtype.
        """
        try:
            dtype = dtype._numpy_dtype
        except Exception:
            pass

        return HDF5ExtensionArray(self._ndarray.astype(dtype), self._column_index, dtype=dtype)

    def take(self, indices, allow_fill=False, fill_value=None):
        from pandas.core.algorithms import take

        # is worth to try to select only a few indices ?
        data = self._ndarray

        if allow_fill and fill_value is None:
            fill_value = self.dtype.na_value

        # fill value should always be translated from the scalar
        # type for the array, to the physical storage type for
        # the data, before passing to take.

        result = take(data, indices, fill_value=fill_value,
                      allow_fill=allow_fill)
        return self._from_sequence(result, dtype=self.dtype)

    def copy(self):
        """
        Return a copy of the array.

        Returns
        -------
        ExtensionArray
        """
        return HDF5ExtensionArray(np.array(self._ndarray))

    def _accumulate(self, name: str, *, skipna: bool = True, **kwargs):
        """
        Return an ExtensionArray performing an accumulation operation.

        The underlying data type might change.

        Parameters
        ----------
        name: str
            Name of the function, supported values are:
            - cummin
            - cummax
            - cumsum
            - cumprod
        skipna: bool, default True
            If True, skip NA values.
        **kwargs
            Additional keyword arguments passed to the accumulation function.
            Currently, there is no supported kwarg.

        Returns
        -------
        array

        Raises
        ------
        NotImplementedError: subclass does not define accumulations
        """
        if skipna:
            meth = getattr(self._ndarray, name, None)
        else:
            array = self._ndarray
            meth = getattr(array[not np.isnan(array)], name, None)
        if meth is None:
            raise TypeError(
                f"'{type(self).__name__}' with dtype {self.dtype} "
                f"does not support reduction '{name}'"
            )
        return meth(**kwargs)

    def _reduce(self, name: str, *, skipna: bool = True, min_count=0, **kwargs):
        """
        Return a scalar result of performing the reduction operation.

        Parameters
        ----------
        name: str
            Name of the function, supported values are:
            {any, all, min, max, sum, mean, median, prod,
            std, var, sem, kurt, skew}.
        skipna: bool, default True
            If True, skip NaN values.
        **kwargs
            Additional keyword arguments passed to the reduction function.
            Currently, `ddof` is the only supported kwarg.

        Returns
        -------
        scalar

        Raises
        ------
        TypeError: subclass does not define reductions
        """
        if skipna:
            meth = getattr(self._ndarray, name, None)
        else:
            array = self._ndarray
            meth = getattr(array[~np.isnan(array)], name, None)
        if meth is None:
            raise TypeError(
                f"'{type(self).__name__}' with dtype {self.dtype} "
                f"does not support reduction '{name}'"
            )
        return meth(**kwargs)

    def __array_ufunc__(self, ufunc: np.ufunc, method: str, *inputs, **kwargs):
        # Lightly modified version of
        # https://numpy.org/doc/stable/reference/generated/numpy.lib.mixins.NDArrayOperatorsMixin.html
        # The primary modification is not boxing scalar return values
        # in NumpyExtensionArray, since pandas' ExtensionArrays are 1-d.
        if any(
            isinstance(other, (ABCSeries, ABCIndex, ABCDataFrame)) for other in inputs
        ):
            return NotImplemented

        # Operations likes cos are much faster with ufunc
        out = kwargs.get("out", ())

        inputs = tuple(inp._ndarray if isinstance(inp, HDF5ExtensionArray) else inp for inp in inputs)

        result = arraylike.maybe_dispatch_ufunc_to_dunder_op(
            self._ndarray, ufunc, method, *inputs, **kwargs
        )
        if result is not NotImplemented:
            if ufunc.nout > 1:
                # multiple return values; re-box array-like results
                return tuple(type(self)(x) for x in result)
            else:
                return type(self)(result)

        if "out" in kwargs:
            # e.g. test_ufunc_unary
            return arraylike.dispatch_ufunc_with_out(
                self._ndarray, ufunc, method, *inputs, **kwargs
            )

        if method == "reduce":
            result = arraylike.dispatch_reduction_ufunc(
                self._ndarray, ufunc, method, *inputs, **kwargs
            )
            if result is not NotImplemented:
                # e.g. tests.series.test_ufunc.TestNumpyReductions
                return type(self)(result)

        # Defer to the implementation of the ufunc on unwrapped values.
        if out:
            kwargs["out"] = tuple(inputs)
        result = getattr(ufunc, method)(*inputs, **kwargs)

        if ufunc.nout > 1:
            # multiple return values; re-box array-like results
            return tuple(type(self)(x) for x in result)
        elif method == "at":
            # no return value
            return None
        elif method == "reduce":
            if isinstance(result, np.ndarray):
                # e.g. test_np_reduce_2d
                return type(self)(result)

            # e.g. test_np_max_nested_tuples
            return type(self)(result)
        else:
            # one return value; re-box array-like results
            return type(self)(result)

    # ------------------------------------------------------------------------
    # Ops

    def value_counts(self, dropna: bool = True) -> Series:
        """
        Return a Series containing counts of unique values.

        Parameters
        ----------
        dropna : bool, optional
            Don't include counts of NaN. The default is True.

        Returns
        -------
        Series
            value_counts.

        """
        values = _ensure_arraylike(self._ndarray, func_name="value_counts")
        keys, counts, _ = value_counts_arraylike(values, dropna)
        if keys.dtype == np.float16:
            keys = keys.astype(np.float32)

        # For backwards compatibility, we let Index do its normal type
        #  inference, _except_ for if if infers from object to bool.
        idx = Index(keys)
        if idx.dtype == bool and keys.dtype == object:
            idx = idx.astype(object)

        result = Series(counts, index=idx, copy=False)
        return result

    def argsort(self):
        """Return for indices of sorted values."""
        return type(self)(np.argsort(self._ndarray))

    def argmin(self):
        """Return for indice of min value."""
        return type(self)(np.argmin(self._ndarray))

    def argmax(self):
        """Return for indice of max value."""
        return type(self)(np.argmax(self._ndarray))

    def dropna(self):
        """Remove missing values."""
        return type(self)(self._ndarray[~np.isna(self._ndarray)])

    def interpolate(
        self,
        *,
        method: InterpolateOptions,
        axis: int,
        index: Index,
        limit,
        limit_direction,
        limit_area,
        copy: bool,
        **kwargs,
    ):
        """
        See NDFrame.interpolate.__doc__.
        """
        # NB: we return type(self) even if copy=False
        if not copy:
            out_data = self._ndarray
        else:
            out_data = self._ndarray.copy()

        # TODO: assert we have floating dtype?
        missing.interpolate_2d_inplace(
            out_data,
            method=method,
            axis=axis,
            index=index,
            limit=limit,
            limit_direction=limit_direction,
            limit_area=limit_area,
            **kwargs,
        )
        if not copy:
            return self
        return type(self)._simple_new(out_data, dtype=self.dtype)

    def memory_usage(self, *args, **kwargs):
        return self.nbytes

    # error: Signature of "__eq__" incompatible with supertype "object"
    def __eq__(self, other: Any) -> ArrayLike:  # type: ignore[override]
        """Return for `self == other` (element-wise equality)."""
        # Implementer note: this should return a boolean numpy ndarray or
        # a boolean ExtensionArray.
        # When `other` is one of Series, Index, or DataFrame, this method should
        # return NotImplemented (to ensure that those objects are responsible for
        # first unpacking the arrays, and then dispatch the operation to the
        # underlying arrays)
        if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
            return NotImplemented
        else:
            return HDF5ExtensionArray(self._ndarray == other)

    # def __invert__(self):
    #     """Return for `~self`."""
    #     return type(self)(~self._ndarray)

    # def __neg__(self):
    #     """Return for `-self`."""
    #     return type(self)(-self._ndarray)

    # def __pos__(self):
    #     """Return for `+self`."""
    #     return type(self)(+self._ndarray)

    # def __abs__(self):
    #     """Return for `abs(self)`."""
    #     return type(self)(abs(self._ndarray))

    # # error: Signature of "__ne__" incompatible with supertype "object"
    # def __ne__(self, other: Any) -> ArrayLike:  # type: ignore[override]
    #     """Return for `self != other` (element-wise in -equality)."""
    #     return type(self)(self._ndarray != other)

    # # error: Signature of "__ne__" incompatible with supertype "object"

    # def __lt__(self, other: Any) -> ArrayLike:  # type: ignore[override]
    #     """Return for `self < other` (element-wise in -equality)."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray < other)

    # def __gt__(self, other: Any) -> ArrayLike:  # type: ignore[override]
    #     """Return for `self < other` (element-wise in -equality)."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray > other)

    # def __le__(self, other: Any) -> ArrayLike:  # type: ignore[override]
    #     """Return for `self < other` (element-wise in -equality)."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray <= other)

    # def __ge__(self, other: Any) -> ArrayLike:  # type: ignore[override]
    #     """Return for `self < other` (element-wise in -equality)."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray >= other)

    # def __add__(self, other: Any) -> ArrayLike:
    #     """Return for `self + other`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray + other)

    # def __radd__(self, other: Any) -> ArrayLike:
    #     """Return for `other + self`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(other + self._ndarray)

    # def __sub__(self, other: Any) -> ArrayLike:
    #     """Return for `self - other`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray - other)

    # def __rsub__(self, other: Any) -> ArrayLike:
    #     """Return for `other - self`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(other - self._ndarray)

    # def __mod__(self, other: Any) -> ArrayLike:
    #     """Return for `self % other`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray % other)

    # def __divmod__(self, other: Any) -> ArrayLike:
    #     """Return for `self % other`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(np.divmod(self, other))

    # def __rdivmod__(self, other: Any) -> ArrayLike:
    #     """Return for `other % self`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(np.divmod(other, self))

    # # def __mul__(self, other: Any) -> ArrayLike:
    # #     """Return for `self * other`."""
    # #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    # #         return NotImplemented
    # #     else:
    # #         return type(self)(self._ndarray * other)

    # def __rmul__(self, other: Any) -> ArrayLike:
    #     """Return for `other * self`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(other * self._ndarray)

    # def __truediv__(self, other: Any) -> ArrayLike:
    #     """Return for `self / other`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray / other)

    # def __floordiv__(self, other: Any) -> ArrayLike:
    #     """Return for `self // other`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray // other)

    # def __rfloordiv__(self, other: Any) -> ArrayLike:
    #     """Return for `other // self`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(other // self._ndarray)

    # def __lshift__(self, value: Any) -> ArrayLike:
    #     """Return for `self << other`."""
    #     if isinstance(value, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray << value)

    # def __rshift__(self, value: Any) -> ArrayLike:
    #     """Return for `self >> other`."""
    #     if isinstance(value, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray >> value)

    # def __and__(self, other: Any) -> ArrayLike:
    #     """Return for `self & other`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray & other)

    # def __rand__(self, other: Any) -> ArrayLike:
    #     """Return for `self & other`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(other & self._ndarray)

    # def __or__(self, other: Any) -> ArrayLike:
    #     """Return for `self | other`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray | other)

    # def __xor__(self, other: Any) -> ArrayLike:
    #     """Return for `self ^ other`."""
    #     if isinstance(other, (pandas.Series, pandas.Index, pandas.DataFrame)):
    #         return NotImplemented
    #     else:
    #         return type(self)(self._ndarray ^ other)

    # Additional functions that are copy of the NumpyExtensionArray functions

    def sem(
        self,
        *,
        axis: AxisInt | None = None,
        dtype: NpDtype | None = None,
        out=None,
        ddof: int = 1,
        keepdims: bool = False,
        skipna: bool = True,
    ):
        nv.validate_stat_ddof_func((), {"dtype": dtype, "out": out, "keepdims": keepdims}, fname="sem")
        result = nanops.nansem(self._ndarray, axis=axis, skipna=skipna, ddof=ddof)
        return self._wrap_reduction_result(axis, result)

    def kurt(
        self,
        *,
        axis: AxisInt | None = None,
        dtype: NpDtype | None = None,
        out=None,
        keepdims: bool = False,
        skipna: bool = True,
    ):
        nv.validate_stat_ddof_func(
            (), {"dtype": dtype, "out": out, "keepdims": keepdims}, fname="kurt"
        )
        result = nanops.nankurt(self._ndarray, axis=axis, skipna=skipna)
        return self._wrap_reduction_result(axis, result)

    def skew(
        self,
        *,
        axis: AxisInt | None = None,
        dtype: NpDtype | None = None,
        out=None,
        keepdims: bool = False,
        skipna: bool = True,
    ):
        nv.validate_stat_ddof_func(
            (), {"dtype": dtype, "out": out, "keepdims": keepdims}, fname="skew"
        )
        result = nanops.nanskew(self._ndarray, axis=axis, skipna=skipna)
        return self._wrap_reduction_result(axis, result)



"""Very thin overlay over h5py library."""

import numpy as np
from h5pandas.dataframe import dataset_to_dataframe, group_to_dataframe
import h5py
try:
    from pandas import DataFrame
except ModuleNotFoundError:
    DataFrame = type(None)


class Group(h5py.Group):
    """
    h5py Group that provides a DataFrame instead of dataset.

    See h5py documentation: https://docs.h5py.org/en/stable/high/group.html
    """

    def __init__(self, group_id, columns=None):
        """
        Transform an object into a Group that provides DataFrame instead of dataset.

        Parameters
        ----------
        group_id : TYPE
            DESCRIPTION.
        columns : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.

        """
        """"""
        if isinstance(group_id, h5py.File):
            id = group_id['/']._id
        elif isinstance(group_id, (h5py.h5f.FileID, h5py.h5g.GroupID)):
            id = group_id
        elif isinstance(group_id, h5py.Group):
            id = group_id._id

        super().__init__(id)

    def __getitem__(self, *args, **kwargs):
        """Convert item into DataFrame before returning it."""
        item = super().__getitem__(*args, **kwargs)
        if isinstance(item, h5py.Group):
            try:
                return group_to_dataframe(item)
            except Exception:
                return Group(item)
        elif isinstance(item, h5py.Dataset):
            try:
                return dataset_to_dataframe(item)
            except Exception:
                return item
        print("Item type not managed")
        return item

    def __getattribute__(self, name):
        """Get DataFrame if possible."""
        item = super().__getattribute__(name)
        if isinstance(item, h5py.File):
            item.__class__ = File
        elif isinstance(item, h5py.Group):
            item = Group(item)
        return item

    def create_dataset(self, name, shape=None, dtype=None, data=None, index=None, columns=None, **kwargs):
        """
        Create a DataFrame.

        If columns is provided or if data is a DataFrame, the columns names are written as attribute of the dataset.

        See h5py documentation: https://docs.h5py.org/en/stable/high/dataset.html

        Parameters
        ----------
        name : TYPE
            DESCRIPTION.
        shape : TYPE, optional
            DESCRIPTION. The default is None.
        dtype : TYPE, optional
            DESCRIPTION. The default is None.
        data : TYPE, optional
            DESCRIPTION. The default is None.
        index : TYPE, optional
            DESCRIPTION. The default is None.
        columns : TYPE, optional
            DESCRIPTION. The default is None.
        kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        pandas.DataFrame
            The newly create DataFrame.

        """
        dataset = super().create_dataset(name, shape=shape, dtype=dtype, data=data, **kwargs)
        # we look for columns name inside the dataFrame
        if isinstance(data, DataFrame) and columns is None:
            columns = list(data.columns)
        # Write columns name inside the dataFrame
        # TODO : decode it with :
        if columns is not None:
            dataset.attrs["columns"] = np.char.encode(np.array(columns).astype(str))

        return dataset_to_dataframe(dataset, index=index, columns=columns)


class File(h5py.File, Group):
    """
    h5py File that provides a DataFrame instead of dataset.

    See h5py documentation: https://docs.h5py.org/en/stable/high/file.html
    """

    def __getitem__(self, *args, **kwargs):
        """Getter of Group class."""
        return super(h5py.File, self).__getitem__(*args, **kwargs)



"""Module related to dataFrames."""
import numpy as np
import pandas
from h5pandas.h5array import HDF5ExtensionArray
import h5py
import pandas


def dataset_to_dataframe(dataset, columns=None, index=None, copy=False):
    """
    Transform a dataset into a DataFrame.

    Parameters
    ----------
    dataset : h5py.dataset
        The dataset to convert into a DataFrame.
    columns : iterable, optional
        Column labels to use for resulting frame when data does not have them,
        defaulting to RangeIndex(0, 1, 2, ..., n). If data contains column labels,
        will perform column selection instead.
    index : Index or array-like, optional
        Index to use for resulting frame. Will default to RangeIndex if
        no indexing information part of input data and no index provided.
    copy : bool, optional
        Copy data from inputs.
        For dict data, the default of None behaves like ``copy=True``.  For DataFrame
        or 2d ndarray input, the default of None behaves like ``copy=False``.
        If data is a dict containing one or more Series (possibly of different dtypes),
        ``copy=False`` will ensure that these inputs are not copied.

    Returns
    -------
    pandas.DataFrame
        A dataFrame backed by the dataset.
        If you change the dataset values, the DataFrame will cbe changed.

    """
    """"""
    # if no columns we try to find columns or we construct a tuple of None
    if columns is None:
        if "columns" in dataset.attrs:
            columns = tuple(np.char.decode(dataset.attrs['columns']))
        else:
            columns = (None,)*dataset.shape[1]

    columns_decoded = [None]*len(columns)
    for i, col in enumerate(columns):
        if isinstance(col, (bytes, np.bytes_)):
            columns_decoded[i] = col.decode()
        else:
            columns_decoded[i] = col

    # we create a Series for each column
    series = (pandas.Series(HDF5ExtensionArray(dataset, i), name=col, copy=False) for i, col in enumerate(columns_decoded))

    # concatenate the series into a DataFrame
    return pandas.concat(series, copy=copy, axis=1)


def group_to_dataframe(group) -> pandas.DataFrame:
    """
    Transform a group into a DataFrame.

    Parameters
    ----------
    group : h5py.group
        The group to convert into a DataFrame.

    Returns
    -------
    pandas.DataFrame
        A dataFrame backed by the dataset.
        If you change the dataset values, the DataFrame will cbe changed.
    """
    # First option : the dataframe has been written by pandas (PyTables) with format = "fixed" or "table"
    if "pandas_type" in group.attrs:
        if group.attrs["pandas_type"] == b"frame":
            return _group_fixed_to_dataframe(group)
        elif group.attrs["pandas_type"] == b"frame_table":
            return _group_table_to_dataframe(group)

    # Second option : all the datasets have the same length, each is one is a serie
    try:
        return _group_with_column_to_dataframe(group)
    except ValueError:
        pass

    raise ValueError("Group could not be converted into a DataFrame")


def _group_with_column_to_dataframe(group) -> pandas.DataFrame:
    series = []
    for dataset_name in group:
        dataset = group[dataset_name]
        if not isinstance(dataset, h5py.dataset):
            raise ValueError("All child of the group must be datasets")
        if "columns" in dataset.attrs:
            raise ValueError("This dataset contains several columns")
        series.append(pandas.Series(HDF5ExtensionArray(dataset), name=dataset_name, copy=False))

    # concatenate the series into a DataFrame
    return pandas.concat(series, axis=1)


def _group_fixed_to_dataframe(group) -> pandas.DataFrame:
    return dataset_to_dataframe(group["block0_values"], columns=group["axis0"], index=group["axis1"])


def _group_table_to_dataframe(group) -> pandas.DataFrame:
    raise NotImplementedError


@pandas.api.extensions.register_dataframe_accessor("h5")
@pandas.api.extensions.register_series_accessor("h5")
class DatasetAccessor:
    """Accessor to dataset for pandas object from h5pandas."""

    def __init__(self, pandas_obj):
        """
        Init the accessor of a Panda Series or DataFrame.

        Parameters
        ----------
        pandas_obj : pandas.Series or pandas.DataFrame

        """
        self._values = self._validate(pandas_obj)
        self._obj = pandas_obj

    @staticmethod
    def _validate(obj):
        """Verify the DataFrame is backed by opened h5file."""
        if isinstance(obj, pandas.DataFrame):
            values = obj[obj.columns[0]].values
        elif isinstance(obj, pandas.Series) and not hasattr(obj.values, "_datatset"):
            values = obj.values
        else:
            values = obj
        if hasattr(values, "_dataset"):
            return values
        else:
            raise AttributeError("Pandas Object must be backed by h5file.")

    @property
    def file(self):
        """Return the file backing the Pandas Object."""
        return self._values._dataset.file

    @property
    def dataset(self):
        """Return the dataset backing the Pandas Object."""
        return self._values._dataset

    @property
    def attrs(self):
        """Return the file backing the Pandas Object."""
        return self._values._dataset.attrs



"""
h5pandas tests.
"""

from pandas.tests.extension import base
import pytest


# class TestConstructors(base.BaseConstructorsTests):
#     # 15 passed/15
#     pass


# class TestCasting(base.BaseCastingTests):
#     # 6 failed, 5 passed, 1 skipped
#     pass


# class TestDtype(base.BaseDtypeTests):
#     # 22 passed
#     def test_construct_from_string_wrong_type_raises(self, dtype):
#         with pytest.raises(
#             TypeError,
#             match="expected string or bytes-like object, got 'int'",
#         ):
#             type(dtype).construct_from_string(0)


# class TestGetitem(base.BaseGetitemTests):
#     # 35 passed, 2 xfailed
#     def test_getitem_invalid(self, data):
#         # TODO: box over scalar, [scalar], (scalar,)?

#         msg = (
#             r"only integers, slices \(`:`\), ellipsis \(`...`\), numpy.newaxis "
#             r"\(`None`\) and integer or boolean arrays are valid indices"
#         )
#         with pytest.raises(IndexError, match=msg):
#             data["foo"]
#         with pytest.raises(IndexError, match=msg):
#             data[2.5]

#         ub = len(data)
#         msg = "|".join(
#             [
#                 "list index out of range",  # json
#                 "index out of bounds",  # pyarrow
#                 "Out of bounds access",  # Sparse
#                 f"loc must be an integer between -{ub} and {ub}",  # Sparse
#                 f"index {ub+1} is out of bounds for axis 0 with size {ub}",
#                 f"index -{ub+1} is out of bounds for axis 0 with size {ub}",
#                 f"Index \({ub+1}\) out of range for \(0-{ub-1}\)",
#                 f"Index \(-1\) out of range for \(0-{ub-1}\)"
#             ]
#         )
#         with pytest.raises(IndexError, match=msg):
#             data[ub + 1]
#         with pytest.raises(IndexError, match=msg):
#             data[-ub - 1]


# class TestGroupby(base.BaseGroupbyTests):
#     # 12 passed
#     pass


# class TestInterface(base.BaseInterfaceTests):
#     #  14 passed
#     pass


# class TestParsing(base.BaseParsingTests):
#     # 2 failed : should be okay
#     pass


# class TestMethods(base.BaseMethodsTests):
#     # 31 failed, 81 passed, 4 errors
#     pass


# class TestMissing(base.BaseMissingTests):
#     # 24 passed
#     pass


# class TestArithmeticOps(base.BaseArithmeticOpsTests):
#     series_scalar_exc: type[Exception] | None = None
#     frame_scalar_exc: type[Exception] | None = None
#     series_array_exc: type[Exception] | None = None
#     divmod_exc: type[Exception] | None = None

#     # 36 failed, 48 passed
#     pass


# class TestComparisonOps(base.BaseComparisonOpsTests):
#     # 12 passed
#     pass


# class TestOpsUtil(base.BaseOpsUtil):
#     # Nothing
#     pass


# class TestUnaryOps(base.BaseUnaryOpsTests):
#     # 4 passed
#     pass


# class TestPrinting(base.BasePrintingTests):
#     # 6 passed
#     pass


# class TestBooleanReduce(base.BaseBooleanReduceTests):
#     # 4 passed, 48 skipped
#     pass


# class TestNoReduce(base.BaseNoReduceTests):
#     # 26 failed, 20 passed, 6 skipped, 16 warnings
#     def _supports_reduction(self, obj, op_name: str) -> bool:
#         # Specify if we expect this reduction to succeed.
#         return True
#     pass


# class TestNumericReduce(base.BaseNumericReduceTests):
#     # 26 failed, 16 passed, 10 skipped, 16 warnings
#     pass


# class TestReshaping(base.BaseReshapingTests):
#     # 15 failed, 17 passed
#     pass


if __name__ == '__main__':
    # retcode = pytest.main(["test_extension.py"])
    retcode = pytest.main(["test_HDFExtensionArray.py"])



"""
Sandbox Tests.
"""

import numpy as np
import h5pandas
from h5pandas import dataset_to_dataframe
from h5pandas import HDF5Dtype
from h5pandas import HDF5ExtensionArray
import h5py
import time
import pandas as pd

HDF5Dtype("i8")


def TestH5extensions():
    arr = np.random.rand(30000, 5)
    with h5py.File("toto2.h5", "w", libver='latest', driver='core', backing_store=False) as f:
        t0 = time.time()
        t00 = t0
        d = f.create_dataset('toto', data=arr)

        # very basic transformation in hdf5array
        array = HDF5ExtensionArray(d, 0)
        d[0, 0] = 1.
        assert (array[0] == 1.)

        # transformation into a series
        ser = pd.Series(array, name='toto', copy=False)
        ser.memory_usage()
        d[0, 0] = 2.
        array[0] = 2.
        assert (ser[0] == 2.)
        assert (d[0, 1] != 2.)

        t1 = time.time()
        print(d.dtype, t1-t0)
        t0 = time.time()

        df = dataset_to_dataframe(d)
        df = dataset_to_dataframe(d, ["a", "b", "c", "d", "e"])
        t1 = time.time()
        print(df, t1-t0)
        # assert on time to make sure the file is not loaded
        assert (t1-t0 < 0.02)
        d[0, 0] = 0
        assert (df.loc[0, 'a'] == 0)
        t0 = time.time()
        b = df["b"]

        ind = b < 0.5
        print(ind, time.time()-t0)
        t0 = time.time()

        b = df["b"]
        sub_df = df[ind]
        mini = df["a"].min()
        cosb = np.cos(df["b"])
        cosb_2 = np.cos(b)
        print(cosb)

        assert all(cosb == cosb_2)
        cosb > cosb_2
        print('Test cos', time.time() - t0)
        t0 = time.time()

        print('Test reduce std')
        res = b.std()
        print('Test std', time.time()-t0)
        t0 = time.time()

        print('Test accumulate cumsum')
        b.cumsum()
        print('Test cumsum', time.time()-t0)
        t0 = time.time()

        # test groupby
        t0 = time.time()
        result = df.groupby("a", as_index=True)
        b = result.b
        b.mean()
        print('Test groupby', time.time()-t0)

        t0 = time.time()
        result = df[(df['a'] < 0.5) & (df['b'] > 0.5)]
        result.c.mean()
        print('test loc :', time.time()-t0)
        t0 = time.time()

        # print(df.h5.file)

        df["z"] = df["a"] + df["b"]*df["c"]
        df["y"] = df["a"] - df["b"]/df["c"]
        df["z"] = df["a"] % df["b"]
        df["z"] = df["a"] % df["b"]
        print('Test basic operations', time.time()-t0)
        t0 = time.time()

        # print(df.h5.attrs.keys())

        # Tester operations with skipna

        # Test hasna

        # Test other types

        # Test casting from a type to another

    # print(b)
    print('Total time : ', time.time() - t00)


def test_rmul():
    arr = np.random.rand(3000, 5)
    with h5py.File("toto2.h5", "w", libver='latest', driver='core', backing_store=False) as f:
        d = f.create_dataset('toto', data=arr)
        df = dataset_to_dataframe(d)

        t0 = time.time()
        df[0]*2.1
        tf = time.time()
        print("rmul = ", tf-t0)

        # t0 = time.time()
        # 3*df[0]
        # tf = time.time()
        print("rmul = ", tf-t0)


def test_op_2EA():
    arr = np.random.rand(3000, 5)
    with h5py.File("toto2.h5", "w", libver='latest', driver='core', backing_store=False) as f:
        d = f.create_dataset('toto', data=arr)
        df = dataset_to_dataframe(d)
        df[0] - df[0]


def test_add():
    arr = np.random.rand(3000, 5)
    with h5py.File("toto2.h5", "w", libver='latest', driver='core', backing_store=False) as f:
        d = f.create_dataset('toto', data=arr)
        df = dataset_to_dataframe(d)

        ser = df[0]
        ser2 = ser + (-57.0)

def TestH5Group():
    arr = np.random.rand(3000, 5)
    df = pd.DataFrame(arr)
    df_named = pd.DataFrame(arr, columns=['é', 'b', 'c', 'd', 'e'], index=range(1000, 4000))

    df.to_hdf("foobar.h5", "random_fixed", format="fixed", mode="w")
    df.to_hdf("foobar.h5", "random_table", format="table")
    df_named.to_hdf("foobar.h5", "named_random_fixed", format="fixed")
    df_named.to_hdf("foobar.h5", "named_random_table", format="table")

    with h5pandas.File("foobar.h5", "a", libver='latest') as f:
        f.create_dataset('h5pandas', data=df)
        f.create_dataset('h5pandas_named', data=df_named)

    with h5pandas.File("foobar.h5", "r", libver='latest') as f:
        df = f['h5pandas']

        df = f['named_random_fixed']


if __name__ == '__main__':
    TestH5extensions()
    TestH5Group()
    test_rmul()
    test_op_2EA()
    test_add()



# Always prefer setuptools over distutils
from setuptools import setup

# To use a consistent encoding
from codecs import open
from os import path

# The directory containing this file
HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# This call to setup() does all the work
setup(
    name="h5pandas",
    version="0.5",
    description="Load hdf5 into Pandas DataFrame instantaneously",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    author="Frédéric MASSON",
    author_email="masson-frederic@hotmail.fr",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent"
    ],
    packages=["h5pandas"],
    include_package_data=True,
    install_requires=["h5py", "pandas>=2.1.0"]
)



"""Pandas ExtensionDataType."""
from pandas.api.extensions import register_extension_dtype, ExtensionDtype
import numpy as np
from pandas._libs.hashtable import object_hash
from pandas.api.extensions import ExtensionArray
from typing import (
    Any,
    TypeVar,
)
from pandas._typing import (
    DtypeObj,
    Shape,
    type_t,
)

# from pandas.core.dtypes.dtypes import PandasDtype
from pandas.core.dtypes.generic import (
    ABCDataFrame,
    ABCIndex,
    ABCSeries,
)

# To parameterize on same ExtensionDtype
ExtensionDtypeT = TypeVar("ExtensionDtypeT", bound="ExtensionDtype")


@register_extension_dtype
class HDF5Dtype(ExtensionDtype):
    """ Ok"""

    _metadata: tuple[str, ...] = ()

    def __init__(self, *args, **kwargs):
        """


        Parameters
        ----------
        *args : TYPE
            DESCRIPTION.
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        dtype = np.dtype(*args, **kwargs)
        self._numpy_dtype = dtype
        self._type = dtype.type
        self._name = dtype.name + "[HDF5]"
        self._metadata = ("name",)
        self._kind = dtype.kind

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: Any) -> bool:
        """
        Check whether 'other' is equal to self.

        By default, 'other' is considered equal if either

        * it's a string matching 'self.name'.
        * it's an instance of this type and all of the attributes
          in ``self._metadata`` are equal between `self` and `other`.

        Parameters
        ----------
        other : Any

        Returns
        -------
        bool
        """
        if isinstance(other, str):
            try:
                other = self.construct_from_string(other)
            except TypeError:
                return False
        if isinstance(other, type(self)):
            return all(
                getattr(self, attr) == getattr(other, attr) for attr in self._metadata
            )
        return False

    def __hash__(self) -> int:
        # for python>=3.10, different nan objects have different hashes
        # we need to avoid that and thus use hash function with old behavior
        return object_hash(tuple(getattr(self, attr) for attr in self._metadata))

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    @property
    def na_value(self) -> object:
        """
        Default NA value to use for this type.

        This is used in e.g. ExtensionArray.take. This should be the
        user-facing "boxed" version of the NA value, not the physical NA value
        for storage.  e.g. for JSONArray, this is an empty dictionary.
        """
        return np.nan

    @property
    def type(self) -> type_t[Any]:
        """
        The scalar type for the array, e.g. ``int``

        It's expected ``ExtensionArray[item]`` returns an instance
        of ``ExtensionDtype.type`` for scalar ``item``, assuming
        that value is valid (not NA). NA values do not need to be
        instances of `type`.
        """
        return self._type

    @property
    def kind(self) -> str:
        """
        A character code (one of 'biufcmMOSUV'), default 'O'

        This should match the NumPy dtype used when the array is
        converted to an ndarray, which is probably 'O' for object if
        the extension type cannot be represented as a built-in NumPy
        type.

        See Also
        --------
        numpy.dtype.kind
        """
        return self._kind

    @property
    def name(self) -> str:
        """
        A string identifying the data type.

        Will be used for display in, e.g. ``Series.dtype``
        """
        return self._name

    @property
    def names(self) -> list[str] | None:
        """
        Ordered list of field names, or None if there are no fields.

        This is for compatibility with NumPy arrays, and may be removed in the
        future.
        """
        return None

    @classmethod
    def construct_array_type(cls) -> type_t[ExtensionArray]:
        """
        Return the array type associated with this dtype.

        Returns
        -------
        type
        """
        from h5pandas import HDF5ExtensionArray
        return HDF5ExtensionArray

    def empty(self, shape: Shape) -> type_t[ExtensionArray]:
        """
        Construct an ExtensionArray of this dtype with the given shape.

        Analogous to numpy.empty.

        Parameters
        ----------
        shape : int or tuple[int]

        Returns
        -------
        ExtensionArray
        """
        cls = self.construct_array_type()
        return cls._empty(shape, dtype=self)

    @classmethod
    def construct_from_string(
        cls: type_t[ExtensionDtypeT], string: str
    ) -> ExtensionDtypeT:
        r"""
        Construct this type from a string.

        This is useful mainly for data types that accept parameters.
        For example, a period dtype accepts a frequency parameter that
        can be set as ``period[H]`` (where H means hourly frequency).

        By default, in the abstract class, just the name of the type is
        expected. But subclasses can overwrite this method to accept
        parameters.

        Parameters
        ----------
        string : str
            The name of the type, for example ``category``.

        Returns
        -------
        ExtensionDtype
            Instance of the dtype.

        Raises
        ------
        TypeError
            If a class cannot be constructed from this 'string'.

        Examples
        --------
        For extension dtypes with arguments the following may be an
        adequate implementation.

        >>> @classmethod
        ... def construct_from_string(cls, string):
        ...     pattern = re.compile(r"^my_type\[(?P<arg_name>.+)\]$")
        ...     match = pattern.match(string)
        ...     if match:
        ...         return cls(**match.groupdict())
        ...     else:
        ...         raise TypeError(
        ...             f"Cannot construct a '{cls.__name__}' from '{string}'"
        ...         )
        """
        import re
        pattern = re.compile(r"^(\w+)\[HDF5\]$")
        match = pattern.match(string)
        if match:
            print("MATCH", match, match[0], match[1])
            print(match.groupdict())
            return cls(match[1])
        else:
            raise TypeError(f"Cannot construct a '{cls.__name__}' from '{string}'")

    @classmethod
    def is_dtype(cls, dtype: object) -> bool:
        """
        Check if we match 'dtype'.

        Parameters
        ----------
        dtype : object
            The object to check.

        Returns
        -------
        bool

        Notes
        -----
        The default implementation is True if

        1. ``cls.construct_from_string(dtype)`` is an instance
           of ``cls``.
        2. ``dtype`` is an object and is an instance of ``cls``
        3. ``dtype`` has a ``dtype`` attribute, and any of the above
           conditions is true for ``dtype.dtype``.
        """
        dtype = getattr(dtype, "dtype", dtype)

        if isinstance(dtype, (ABCSeries, ABCIndex, ABCDataFrame, np.dtype)):
            # https://github.com/pandas-dev/pandas/issues/22960
            # avoid passing data to `construct_from_string`. This could
            # cause a FutureWarning from numpy about failing elementwise
            # comparison from, e.g., comparing DataFrame == 'category'.
            return False
        elif dtype is None:
            return False
        elif isinstance(dtype, cls):
            return True
        if isinstance(dtype, str):
            try:
                return cls.construct_from_string(dtype) is not None
            except TypeError:
                return False
        return False

    @property
    def _is_numeric(self) -> bool:
        """
        Whether columns with this dtype should be considered numeric.

        By default ExtensionDtypes are assumed to be non-numeric.
        They'll be excluded from operations that exclude non-numeric
        columns, like (groupby) reductions, plotting, etc.
        """
        return np.issubdtype(self._numpy_dtype, np.number)

    @property
    def _is_boolean(self) -> bool:
        """
        Whether this dtype should be considered boolean.

        By default, ExtensionDtypes are assumed to be non-numeric.
        Setting this to True will affect the behavior of several places,
        e.g.

        * is_bool
        * boolean indexing

        Returns
        -------
        bool
        """
        return self._numpy_dtype == np.dtype('bool')

    def _get_common_dtype(self, dtypes: list[DtypeObj]) -> DtypeObj | None:
        """
        Return the common dtype, if one exists.

        Used in `find_common_type` implementation. This is for example used
        to determine the resulting dtype in a concat operation.

        If no common dtype exists, return None (which gives the other dtypes
        the chance to determine a common dtype). If all dtypes in the list
        return None, then the common dtype will be "object" dtype (this means
        it is never needed to return "object" dtype from this method itself).

        Parameters
        ----------
        dtypes : list of dtypes
            The dtypes for which to determine a common dtype. This is a list
            of np.dtype or ExtensionDtype instances.

        Returns
        -------
        Common dtype (np.dtype or ExtensionDtype) or None
        """
        if len(set(dtypes)) == 1:
            # only itself
            return self
        else:
            return None

    @property
    def _can_hold_na(self) -> bool:
        """
        Can arrays of this dtype hold NA values?
        """
        return True
